import mindspore.nn as nn
import mindspore.ops as ops
from mindspore.common.initializer import Normal
from mindspore.nn import Cell


def _conv(in_channels, out_channels, kernel_size=3, stride=1, padding=0, pad_mode='same', has_bias=False):
    init_value = Normal()
    return nn.Conv2d(in_channels, out_channels,
                     kernel_size=kernel_size, stride=stride, padding=padding,
                     pad_mode=pad_mode, weight_init=init_value, has_bias=has_bias)


class FCEHead(Cell):
    """The class for implementing FCENet head.
    FCENet(CVPR2021): Fourier Contour Embedding for Arbitrary-shaped Text
    Detection.

    [https://arxiv.org/abs/2104.10442]

    Args:
        in_channels (int): The number of input channels.
        fourier_degree (int) : The maximum Fourier transform degree k.
    """

    def __init__(self, in_channels, fourier_degree=5):
        super(FCEHead, self).__init__()
        assert isinstance(in_channels, int)

        self.downsample_ratio = 1.0
        self.in_channels = in_channels
        self.fourier_degree = fourier_degree
        self.out_channels_cls = 4
        self.out_channels_reg = (2 * self.fourier_degree + 1) * 2
        self.softmax = nn.Softmax(axis=1)

        self.out_conv_cls = _conv(self.in_channels, self.out_channels_cls, kernel_size=3, padding=1, pad_mode='pad',
                                  has_bias=True)

        self.out_conv_reg = _conv(self.in_channels, self.out_channels_reg, kernel_size=3, padding=1, pad_mode='pad',
                                  has_bias=True)

    def construct(self, feature):
        cls_res = [self.out_conv_cls(feature[0]), self.out_conv_cls(feature[1]), self.out_conv_cls(feature[2])]
        reg_res = [self.out_conv_reg(feature[0]), self.out_conv_reg(feature[1]), self.out_conv_reg(feature[2])]
        level_num = len(cls_res)

        if not self.training:
            preds = []
            for i in range(level_num):
                tr_pred = self.softmax(cls_res[i][:, 0:2, :, :])
                tcl_pred = self.softmax(cls_res[i][:, 2:, :, :])
                preds.append(ops.concat([tr_pred, tcl_pred, reg_res[i]], axis=1))
        else:
            preds = [[cls_res[0], reg_res[0]], [cls_res[1], reg_res[1]], [cls_res[2], reg_res[2]]]

        return [preds[0], preds[1], preds[2]]
